(function(tree) {

tree.Call = function Call(name, args, index) {
    this.is = 'call';

    this.name = name;
    this.args = args;
    this.index = index;
};

tree.Call.prototype = {
    //
    // When evaluating a function call,
    // we either find the function in `tree.functions` [1],
    // in which case we call it, passing the  evaluated arguments,
    // or we simply print it out as it appeared originally [2].
    //
    // The *functions.js* file contains the built-in functions.
    //
    // The reason why we evaluate the arguments, is in the case where
    // we try to pass a variable to a function, like: `saturate(@color)`.
    // The function should receive the value, not the variable.
    //
    eval: function(env) {
        var args = this.args.map(function(a) { return a.eval(env); });

        for (var i = 0; i < args.length; i++) {
            if (args[i].is === 'undefined') {
                return {
                    is: 'undefined',
                    value: 'undefined'
                };
            }
        }

        if (this.name in tree.functions) {
            if (tree.functions[this.name].length === args.length) {
                return tree.functions[this.name].apply(tree.functions, args);
            } else {
                env.error({
                    message: 'incorrect number of arguments for ' + this.name +
                        '(). ' + tree.functions[this.name].length + ' expected.',
                    index: this.index,
                    type: 'runtime',
                    filename: this.filename
                });
                return {
                    is: 'undefined',
                    value: 'undefined'
                };
            }
        } else {
            var fn = tree.Reference.mapnikFunction(this.name);
            if (!fn) {
                env.error({
                    message: 'unknown function ' + this.name,
                    index: this.index,
                    type: 'runtime',
                    filename: this.filename
                });
                return {
                    is: 'undefined',
                    value: 'undefined'
                };
            }
            if (fn[1] !== args.length) {
                env.error({
                    message: 'function ' + this.name + ' takes ' +
                        fn[1] + ' arguments and was given ' + args.length,
                    index: this.index,
                    type: 'runtime',
                    filename: this.filename
                });
                return {
                    is: 'undefined',
                    value: 'undefined'
                };
            } else {
                // Save the evaluated versions of arguments
                this.args = args;
                return this;
            }
        }
    },

    toString: function(env, format) {
        if (format === 'image-filter') {
            if (this.args.length) {
                return this.name + ':' + this.args.join(',');
            } else {
                return this.name;
            }
        } else {
            if (this.args.length) {
                return this.name + '(' + this.args.join(',') + ')';
            } else {
                return this.name;
            }
        }
    }
};

})(require('../tree'));
